import grok
from zope import component

from zope.security import checkPermission

from horae.core import utils

from horae.properties import interfaces


class Properties(grok.OrderedContainer):
    """ Base implementation of a property container
    """
    grok.baseclass()
    grok.implements(interfaces.IProperties)

    # To be defined in the subclass
    holder_interface = None
    interface = None
    default_property_interface = None

    def available(self, property, initial=None, editable=None, display=None):
        """ Whether the provided property is available
        """
        return ((initial is None or property.initial == initial) and
                (editable is None or property.editable == editable) and
                (display is None or property.display == display) and
                (property.permission is None or checkPermission(property.permission, self)))

    def filter(self, properties, initial=None, editable=None, display=None, ignore=[]):
        """ Filter a list of properties
        """
        filtered = []
        for property in properties:
            if not self.available(property, initial, editable, display) or property.id in ignore:
                continue
            filtered.append(property)
        return filtered

    def properties(self, initial=None, editable=None, display=None):
        """ Returns the available properties
        """
        parent = utils.findParentByInterface(self, self.holder_interface, 1)
        properties = self.filter(self.values(), initial, editable, display)
        if parent is not None:
            properties += self.filter(self.interface(parent).properties(), initial, editable, display, self)
        else:
            properties += self.filter(component.getAllUtilitiesRegisteredFor(self.default_property_interface), initial, editable, display, self)
        return sorted(properties, key=lambda property: property.order)

    def add_property(self, property):
        """ Adds a property
        """
        self[property.id] = property

    def del_property(self, id):
        """ Deletes the property with the specified id
        """
        del self[id]

    def get_property(self, id):
        """ Returns the property with the specified id
        """
        if not id in self:
            parent = utils.findParentByInterface(self, self.holder_interface, 1)
            if parent is not None:
                return self.interface(parent).get_property(id)
            else:
                return component.queryUtility(self.default_property_interface, name=id)
        return self[id]

    def fields(self, initial=None, editable=None, display=None):
        """ Iterator over the fields of the available properties
        """
        for property in self.properties(initial, editable, display):
            for field in property.fields():
                yield field


class GlobalProperties(Properties):
    """ Holds properties available for all objects
    """
    grok.implements(interfaces.IGlobalProperties)
    holder_interface = interfaces.IGlobalPropertiesHolder
    interface = interfaces.IGlobalProperties
    default_property_interface = interfaces.IDefaultGlobalProperty


@grok.adapter(interfaces.IGlobalPropertiesHolder)
@grok.implementer(interfaces.IGlobalProperties)
def global_properties_of_holder(holder):
    """ Returns a container for global properties if it does not yet
        exist one is created
    """
    if not 'global_properties' in holder:
        holder['global_properties'] = GlobalProperties()
    return holder['global_properties']


@grok.adapter(interfaces.IPropertied, name='global')
@grok.implementer(interfaces.IProperties)
def global_properties_for_propertied(propertied):
    """ Provides global properties for :py:class:`horae.properties.interfaces.IPropertied`
    """
    holder = utils.findParentByInterface(propertied, interfaces.IGlobalPropertiesHolder)
    return interfaces.IGlobalProperties(holder)


class ClientProperties(Properties):
    """ Holds properties for clients
    """
    grok.implements(interfaces.IClientProperties)
    holder_interface = interfaces.IClientPropertiesHolder
    interface = interfaces.IClientProperties
    default_property_interface = interfaces.IDefaultClientProperty


@grok.adapter(interfaces.IClientPropertiesHolder)
@grok.implementer(interfaces.IClientProperties)
def client_properties_of_holder(holder):
    """ Returns a container for client properties if it does not yet
        exist one is created
    """
    if not 'client_properties' in holder:
        holder['client_properties'] = ClientProperties()
    return holder['client_properties']


class ProjectProperties(Properties):
    """ Holds properties for projects
    """
    grok.implements(interfaces.IProjectProperties)
    holder_interface = interfaces.IProjectPropertiesHolder
    interface = interfaces.IProjectProperties
    default_property_interface = interfaces.IDefaultProjectProperty


@grok.adapter(interfaces.IProjectPropertiesHolder)
@grok.implementer(interfaces.IProjectProperties)
def project_properties_of_holder(holder):
    """ Returns a container for project properties if it does not yet
        exist one is created
    """
    if not 'project_properties' in holder:
        holder['project_properties'] = ProjectProperties()
    return holder['project_properties']


class MilestoneProperties(Properties):
    """ Holds properties for milestones
    """
    grok.implements(interfaces.IMilestoneProperties)
    holder_interface = interfaces.IMilestonePropertiesHolder
    interface = interfaces.IMilestoneProperties
    default_property_interface = interfaces.IDefaultMilestoneProperty


@grok.adapter(interfaces.IMilestonePropertiesHolder)
@grok.implementer(interfaces.IMilestoneProperties)
def milestone_properties_of_holder(holder):
    """ Returns a container for milestone properties if it does not yet
        exist one is created
    """
    if not 'milestone_properties' in holder:
        holder['milestone_properties'] = MilestoneProperties()
    return holder['milestone_properties']


class TicketProperties(Properties):
    """ Holds properties for tickets
    """
    grok.implements(interfaces.ITicketProperties)
    holder_interface = interfaces.ITicketPropertiesHolder
    interface = interfaces.ITicketProperties
    default_property_interface = interfaces.IDefaultTicketProperty


@grok.adapter(interfaces.ITicketPropertiesHolder)
@grok.implementer(interfaces.ITicketProperties)
def ticket_properties_of_holder(holder):
    """ Return a container for milestone properties if it does not yet
        exist one is created
    """
    if not 'ticket_properties' in holder:
        holder['ticket_properties'] = TicketProperties()
    return holder['ticket_properties']



from zope import interface
from zope import schema
from zope.schema.interfaces import IField
from z3c.relationfield.interfaces import IHasRelations

from horae.core import interfaces
from horae.lifecycle.interfaces import ILifecycleAware

from horae.properties import _


class IProperties(interface.Interface):
    """ Property container
    """

    def available(property, initial=None, editable=None, display=None):
        """ Whether the provided property is available
        """

    def filter(properties, initial=None, editable=None, display=None, ignore=[]):
        """ Filter a list of properties
        """

    def properties(initial=None, editable=None, display=None):
        """ Returns the available properties
        """

    def add_property(property):
        """ Adds a property
        """

    def del_property(id):
        """ Deletes the property with the specified id
        """

    def get_property(id):
        """ Returns the property with the specified id
        """

    def fields(initial=None, editable=None, display=None):
        """ Iterator over the fields of the available properties
        """


class IPropertiesHolder(interface.Interface):
    """ Marker interface for objects holding properties
    """


class IGlobalProperties(IProperties):
    """ Holds properties available for all objects
    """


class IClientProperties(IProperties):
    """ Holds properties for clients
    """


class IProjectProperties(IProperties):
    """ Holds properties for projects
    """


class IMilestoneProperties(IProperties):
    """ Holds properties for milestones
    """


class ITicketProperties(IProperties):
    """ Holds properties for tickets
    """


class IGlobalPropertiesHolder(IPropertiesHolder):
    """ Marker interface for objects holding global properties
    """


class IClientPropertiesHolder(IPropertiesHolder):
    """ Marker interface for objects holding client properties
    """


class IProjectPropertiesHolder(IPropertiesHolder):
    """ Marker interface for objects holding project properties
    """


class IMilestonePropertiesHolder(IPropertiesHolder):
    """ Marker interface for objects holding milestone properties
    """


class ITicketPropertiesHolder(IPropertiesHolder):
    """ Marker interface for objects holding ticket properties
    """


class IPropertyTypes(interface.Interface):
    """ Provider for available property types
    """

    def types():
        """ Returns a list of available property types
        """


class IPropertyField(IField):
    """ A field of a property
    """

    property = interface.Attribute('property', 'The property this field belongs to')


class IProperty(interfaces.ITextId):
    """ A property
    """

    type = interface.Attribute('type',
        """ Type of property
        """
    )

    name = schema.TextLine(
        title=_(u'Name'),
        required=True
    )

    description = schema.Text(
        title=_(u'Description'),
        required=False
    )

    required = schema.Bool(
        title=_(u'Required'),
        required=False,
        default=False
    )

    initial = schema.Bool(
        title=_(u'Initial'),
        description=_(u'Whether a property is editable while creating an object'),
        required=False,
        default=True
    )

    editable = schema.Bool(
        title=_(u'Editable'),
        description=_(u'Whether a property is editable on subsequent edits'),
        required=False,
        default=True
    )

    order = schema.Int(
        title=_(u'Order'),
        description=_(u'An integer which defines the order of the different properties'),
        required=True,
        default=1
    )

    display = schema.Bool(
        title=_(u'Display this property in the view of the object'),
        required=False,
        default=True
    )

    remember = schema.Bool(
        title=_(u'Remember the properties value for subsequent edits'),
        required=False,
        default=True
    )

    searchable = schema.Bool(
        title=_(u'Whether users may search for values of this field'),
        required=False,
        default=True
    )

    complete = schema.Bool(
        title=_(u'Whether this field is available for objects having a completed state'),
        required=False,
        default=True
    )

    offer = schema.Bool(
        title=_(u'Whether this field is available for objects having an offer state'),
        required=False,
        default=True
    )

    customizable = interface.Attribute('customizable')
    permission = interface.Attribute('permission')

    def custom_widget(form_fields):
        """ Called after the form fields for the add or edit form of a property have been created
            may be used to add custom widgets to specific fields
        """

    def fields(context):
        """ Returns a list of fields (IPropertyField)
        """

    def field_names():
        """ Returns a list of field names provided by this property
        """

    def process(**data):
        """ Processes the input coming from the form and returns a new data dictionary
        """

    def apply(obj, **data):
        """ Applies the data to the obj
        """

    def render(value, context, request, widget=None):
        """ Returns the rendered value
        """

    def clone():
        """ Returns a copy of the property
        """

    def validate(data, context):
        """ Validates the form data

            Returns a list of exceptions or an empty one if the data is valid
        """

    def empty(value):
        """ Decides whether the value provided is an empty value for this property type
        """


class IDefaultProperty(interface.Interface):
    """ Marker interface for default undeletable properties
    """


class IDefaultGlobalProperty(IDefaultProperty):
    """ Marker interface for default undeletable global properties
    """


class IDefaultClientProperty(IDefaultProperty):
    """ Marker interface for default undeletable client properties
    """


class IDefaultProjectProperty(IDefaultProperty):
    """ Marker interface for default undeletable project properties
    """


class IDefaultMilestoneProperty(IDefaultProperty):
    """ Marker interface for default undeletable global properties
    """


class IDefaultTicketProperty(IDefaultProperty):
    """ Marker interface for default undeletable ticket properties
    """


class IBoolProperty(IProperty):
    """ A boolean property
    """

    default = schema.Bool(
        title=_(u'Default value'),
        required=False
    )


class ITextLineProperty(IProperty):
    """ A textline property
    """

    default = schema.TextLine(
        title=_(u'Default value'),
        required=False
    )


class ITextProperty(IProperty):
    """ A text property
    """

    default = schema.Text(
        title=_(u'Default value'),
        required=False
    )


class IRichTextProperty(IProperty):
    """ A rich text property
    """


class IChoiceProperty(IProperty):
    """ A choice property
    """

    vocabulary = schema.List(
        title=_(u'Options'),
        description=_(u'Specify the available options'),
        required=True,
        value_type=schema.TextLine()
    )

    default = schema.TextLine(
        title=_(u'Default value'),
        required=False
    )

    def getVocabulary(context):
        """ Returns the vocabulary used by the field
        """

    @interface.invariant
    def defaultInVocabulary(choice_property):
        if not choice_property.default in choice_property.vocabulary:
            raise interface.Invalid(_(u'Default value not part of the options'))


class IMultipleChoiceProperty(IChoiceProperty):
    """ A multiple choice property
    """

    default = schema.List(
        title=_(u'Default value'),
        required=False,
        value_type=schema.TextLine()
    )

    @interface.invariant
    def defaultInVocabulary(multiple_choice_property):
        for default in multiple_choice_property.default:
            if not default in multiple_choice_property.vocabulary:
                raise interface.Invalid(_(u'Default value not part of the options'))


class IWeightedOption(interface.Interface):
    """ A weighted option
    """

    name = schema.TextLine(
        title=_(u'Name'),
        required=True
    )

    weight = schema.Int(
        title=_(u'Weight'),
        required=True,
        default=1
    )


class IWeightedChoiceProperty(IChoiceProperty):
    """ A choice property having weighted options
    """

    vocabulary = schema.List(
        title=_(u'Options'),
        description=_(u'Specify the available options'),
        required=True,
        value_type=schema.Object(
            schema=IWeightedOption
        )
    )

    affects_sorting = schema.Bool(
        title=_(u'Affects sorting'),
        description=_(u'Whether this property affects sorting of the object or not')
    )

    @interface.invariant
    def defaultInVocabulary(weighted_choice_property):
        if not weighted_choice_property.default in [option.name for option in weighted_choice_property.vocabulary]:
            raise interface.Invalid(_(u'Default value not part of the options'))


class IFloatProperty(IProperty):
    """ A float property
    """

    default = schema.Float(
        title=_(u'Default value'),
        required=False
    )

    min = schema.Float(
        title=_(u'Minimum value'),
        required=False
    )

    max = schema.Float(
        title=_(u'Maximum value'),
        required=False
    )


class IPriceProperty(IFloatProperty):
    """ A price property
    """


class IUserProperty(IProperty):
    """ A user property
    """

    default = schema.Choice(
        title=_(u'Default value'),
        required=False,
        vocabulary='horae.auth.vocabulary.usernames'
    )


class IUserRoleProperty(IUserProperty):
    """ A user role property
    """

    role = schema.Choice(
        title=_(u'Role'),
        required=True,
        vocabulary='horae.auth.vocabulary.selectableroles'
    )


class IGroupProperty(IProperty):
    """ A group property
    """

    default = schema.Choice(
        title=_(u'Default value'),
        required=False,
        vocabulary='horae.auth.vocabulary.groupids'
    )


class IGroupRoleProperty(IGroupProperty):
    """ A group role property
    """

    role = schema.Choice(
        title=_(u'Role'),
        required=True,
        vocabulary='horae.auth.vocabulary.selectableroles'
    )


class IDateTimeProperty(IProperty):
    """ A date time property
    """

    default = schema.Datetime(
        title=_(u'Default value'),
        required=False
    )

    default_now = schema.Bool(
        title=_(u'Defaults to the date and time when viewed'),
        required=False,
        default=True
    )


class IDateTimeRangeProperty(IDateTimeProperty):
    """ A date time range property
    """

    name = schema.TextLine(
        title=_(u'Name (start date)'),
        required=True
    )

    description = schema.Text(
        title=_(u'Description (start date)'),
        required=False
    )

    required = schema.Bool(
        title=_(u'Required (start date)'),
        required=False
    )

    name_end = schema.TextLine(
        title=_(u'Name (end date)'),
        required=True
    )

    description_end = schema.Text(
        title=_(u'Description (end date)'),
        required=False
    )

    required_end = schema.Bool(
        title=_(u'Required (end date)'),
        required=False,
        default=False
    )

    default_diff = schema.Float(
        title=_(u'The default difference (in hours) between start and end date'),
        required=False,
        default=0.0
    )

    hours = schema.Bool(
        title=_(u'Show an additional hours field representing the range in hours'),
        required=False,
        default=False
    )


class IMilestoneProperty(IProperty):
    """ A milestone property
    """

    default = schema.Choice(
        title=_(u'Default value'),
        required=False,
        vocabulary='horae.properties.vocabulary.milestones'
    )


class IFieldsProperty(IProperty):
    """ A fields property
    """


class IPropertied(interface.Interface):
    """ An object storing dynamic properties and providing a history over them
    """

    def properties():
        """ Iterator over the available properties (IProperty)
        """

    def find_property(id):
        """ Trys to find and return the property having the specified id
        """

    def new_change():
        """ Starts a new property change and returns it
        """

    def get_property(id, default=None):
        """ Return the specific property value
        """

    def set_property(id, value):
        """ Sets a property on the currently active property change
        """

    def changelog():
        """ Iterator over the history of property changes
        """

    def current():
        """ Returns the current property change
        """


class IPropertiesProxy(interface.Interface):
    """ Proxy around IPropertied
    """

    def __getattr__(name):
        """ Returns the current property value
        """

    def __setattr__(name, value):
        """ Sets the value to the current property change
        """


class IPropertyChange(interfaces.IIntId, ILifecycleAware, IHasRelations):
    """ A change of multiple properties
    """

    def get_property(name, default=None):
        """ Returns the value of the specified property
        """

    def set_property(name, value):
        """ Sets a property to be changed
        """

    def properties():
        """ Iterator over the changed properties (name, value)
        """


class IPropertiedDisplayWidgetsProvider(interface.Interface):
    """ An adapter modifying and/or extending the widgets of a propertieds display view
    """

    def widgets(widgets, request):
        """ Returns a modified or extended list of widgets to be displayed
        """


class IHistoryPropertiesProvider(interface.Interface):
    """ An adapter providing properties for the history view
    """

    def properties(change, properties, request):
        """ Returns a modified or extended list of properties for the provided property change
        """


class IComplete(interface.Interface):
    """ Adapter to determine whether an object has been completed or not
    """

    def __call__():
        """ Returns whether the object has been completed or not
        """


class IOffer(interface.Interface):
    """ Adapter to determine whether an object is in the offer phase or not
    """

    def __call__():
        """ Returns whether the object is in the offer phase or not
        """


class IObjectType(interface.Interface):
    """ Adapter to define the user readable object type
    """

    def __call__():
        """ Returns the user readable object type
        """



import grok
import os

from grokcore.chameleon.components import ChameleonPageTemplate
from zope import component
from zope.i18n import translate
from zope.formlib.widget import renderElement
from megrok import navigation

from horae.core import utils
from horae.core.interfaces import ITextIdManager
from horae.layout.interfaces import IActionsMenu, IViewsMenu, IContextualManageMenu
from horae.layout import layout
from horae.layout import _ as _l
from horae.auth.utils import getUser
from horae.auth.interfaces import IUserURL

from horae.properties import _
from horae.properties import interfaces

grok.templatedir('templates')


class DeletePropertied(layout.DeleteForm):
    """ Delete form of a propertied
    """
    grok.context(interfaces.IPropertied)
    grok.require('horae.Delete')
    grok.name('delete')
    navigation.menuitem(IActionsMenu, _(u'Delete'))

    def item_title(self):
        return getattr(self.context, 'name', self.context.id)

    def object_type(self):
        return interfaces.IObjectType(self.context, lambda: _l(u'Item'))()


class PropertiesOverview(layout.View):
    """ Overview of available property containers
    """
    grok.context(interfaces.IPropertiesHolder)
    grok.require('horae.Manage')
    grok.name('properties')
    navigation.menuitem(IContextualManageMenu, _(u'Properties'))

    def update(self):
        super(PropertiesOverview, self).update()
        self.properties = []
        if interfaces.IGlobalPropertiesHolder.providedBy(self.context):
            self.properties.append({'title': _(u'Global properties'),
                                    'description': _(u'Properties available for all objects'),
                                    'url': self.url(interfaces.IGlobalProperties(self.context))})
        if interfaces.IClientPropertiesHolder.providedBy(self.context):
            self.properties.append({'title': _(u'Client properties'),
                                    'description': _(u'Properties available for clients'),
                                    'url': self.url(interfaces.IClientProperties(self.context))})
        if interfaces.IProjectPropertiesHolder.providedBy(self.context):
            self.properties.append({'title': _(u'Project properties'),
                                    'description': _(u'Properties available for projects'),
                                    'url': self.url(interfaces.IProjectProperties(self.context))})
        if interfaces.IMilestonePropertiesHolder.providedBy(self.context):
            self.properties.append({'title': _(u'Milestone properties'),
                                    'description': _(u'Properties available for milestones'),
                                    'url': self.url(interfaces.IMilestoneProperties(self.context))})
        if interfaces.ITicketPropertiesHolder.providedBy(self.context):
            self.properties.append({'title': _(u'Ticket properties'),
                                    'description': _(u'Properties available for tickets'),
                                    'url': self.url(interfaces.ITicketProperties(self.context))})


class Properties(layout.View):
    """ Overview of properties
    """
    grok.context(interfaces.IProperties)
    grok.name('index')
    grok.require('horae.Manage')
    grok.template('properties')

    button_template = u'<a href="%(url)s" class="button button-discreet %(cls)s">%(label)s</a> '

    def label(self):
        if interfaces.IGlobalProperties.providedBy(self.context):
            return _(u'Global properties')
        if interfaces.IClientProperties.providedBy(self.context):
            return _(u'Client properties')
        if interfaces.IProjectProperties.providedBy(self.context):
            return _(u'Project properties')
        if interfaces.IMilestoneProperties.providedBy(self.context):
            return _(u'Milestone properties')
        if interfaces.ITicketProperties.providedBy(self.context):
            return _(u'Ticket properties')

    def get_table(self, properties):
        table = component.getMultiAdapter((self.context, self.request), name='table')
        table.page_size = None
        table.columns = [('name', _(u'Name')), ('description', _(u'Description')), ('type', _(u'Type')), ('actions', u'')]
        table.sortable = {}
        table.base_url = self.url(self.context)
        table.rows = properties
        return table()

    def update(self):
        super(Properties, self).update()
        self.back = self.url(self.context.__parent__) + '/properties'
        properties = self.context.properties()
        self.custom_properties = []
        self.inherited_properties = []
        for property in properties:
            item = {'name': property.name,
                    'description': property.description,
                    'type': property.type,
                    'actions': ''}
            if not interfaces.IDefaultProperty.providedBy(property) and property.__parent__ == self.context:
                item['actions'] = '<div class="button-group">%s</div>' % (
                                    (self.button_template % {'url': self.url(self.context) + '/edit-property?id=' + property.id,
                                                             'label': translate(_(u'Edit'), context=self.request),
                                                             'cls': ''}) + \
                                    (self.button_template % {'url': self.url(self.context) + '/delete-property?id=' + property.id,
                                                             'label': translate(_(u'Delete'), context=self.request),
                                                             'cls': 'button-destructive delete'}))
                self.custom_properties.append(item)
            else:
                if property.customizable:
                    item['actions'] = '<div class="button-group">%s</div>' % \
                                        self.button_template % {'url': self.url(self.context) + '/customize-property?id=' + property.id,
                                                                'label': translate(_(u'Customize'), context=self.request),
                                                                'cls': ''}
                self.inherited_properties.append(item)
        if self.custom_properties:
            self.custom_properties = self.get_table(self.custom_properties)
        if self.inherited_properties:
            self.inherited_properties = self.get_table(self.inherited_properties)
        normalizer = component.getUtility(ITextIdManager)
        types = component.getUtility(interfaces.IPropertyTypes).types()
        self.types = [dict(value=normalizer.normalize(type.type),
                           title=type.type) for type in types]


class PropertyFormMixin(object):
    """ Mixin class for forms of :py:class:`horae.properties.interfaces.IProperty`
    """

    def next_url(self, obj=None):
        return self.cancel_url()


class AddProperty(PropertyFormMixin, layout.AddForm):
    """ Add form for a :py:class:`horae.properties.interfaces.IProperty`
    """
    grok.context(interfaces.IProperties)
    grok.require('horae.Manage')
    grok.name('add-property')

    def __call__(self, type=None):
        self.type = self._findType(type)
        if self.type is not None:
            self.form_fields = grok.AutoFields(self.type).omit('id')
            self.type().custom_widget(self.form_fields)
            normalizer = component.getUtility(ITextIdManager)
            self.additional = renderElement('input',
                                            type='hidden',
                                            name='type',
                                            value=normalizer.normalize(self.type.type))
        else:
            self.redirect(self.cancel_url())
        return super(AddProperty, self).__call__()

    def _findType(self, type):
        if type is None:
            return None
        normalizer = component.getUtility(ITextIdManager)
        types = component.getUtility(interfaces.IPropertyTypes).types()
        for t in types:
            if normalizer.normalize(t.type) == type:
                return t

    def object_type(self):
        return self.type.type

    def create(self, **data):
        property = self.type()
        property.id = component.getUtility(ITextIdManager).idFromName(self.context, data['name'])
        return property

    def add(self, obj):
        self.context.add_property(obj)


class EditProperty(PropertyFormMixin, layout.EditForm):
    """ Edit form for a :py:class:`horae.properties.interfaces.IProperty`
    """
    grok.context(interfaces.IProperties)
    grok.require('horae.Manage')
    grok.name('edit-property')

    def __call__(self, id=None):
        self.property = self.context.get_property(id)
        if self.property is not None:
            self.form_fields = grok.AutoFields(self.property.__class__).omit('id')
            self.property.custom_widget(self.form_fields)
            self.additional = renderElement('input',
                                            type='hidden',
                                            name='id',
                                            value=self.property.id)
        else:
            self.redirect(self.cancel_url())
        return super(EditProperty, self).__call__()

    def object_type(self):
        return self.property.type

    def setUpWidgets(self, ignore_request=False):
        self.context, context = self.property, self.context
        super(EditProperty, self).setUpWidgets(ignore_request)
        self.context = context

    def apply(self, **data):
        self.applyData(self.context, **data)


class CustomizeProperty(PropertyFormMixin, layout.AddForm):
    """ Customization view for an existing
        :py:class:`horae.properties.interfaces.IProperty`
    """
    grok.context(interfaces.IProperties)
    grok.require('horae.Manage')
    grok.name('customize-property')

    def __call__(self, id=None):
        self.property = self.context.get_property(id)
        if self.property is None:
            self.redirect(self.cancel_url())
            return ''
        self.form_fields = grok.AutoFields(self.property.__class__).omit('id')
        self.property.custom_widget(self.form_fields)
        self.additional = renderElement('input',
                                        type='hidden',
                                        name='id',
                                        value=self.property.id)
        return super(CustomizeProperty, self).__call__()

    def object_type(self):
        return self.property.type

    def create(self, **data):
        return self.property.clone()

    def setUpWidgets(self, ignore_request=False):
        super(CustomizeProperty, self).setUpWidgets(ignore_request)
        for field in self.form_fields:
            name = field.field.__name__
            if ignore_request or not self.widgets[name].hasInput():
                self.widgets[name].setRenderedValue(getattr(self.property, name, None))

    def add(self, obj):
        self.context.add_property(obj)


class DeleteProperty(PropertyFormMixin, layout.DeleteForm):
    """ Delete form for a :py:class:`horae.properties.interfaces.IProperty`
    """
    grok.context(interfaces.IProperties)
    grok.require('horae.Manage')
    grok.name('delete-property')

    def __call__(self, id=None):
        self.property = self.context.get_property(id)
        if self.property is not None:
            self.additional = renderElement('input',
                                            type='hidden',
                                            name='id',
                                            value=self.property.id)
        else:
            self.redirect(self.cancel_url())
        return super(DeleteProperty, self).__call__()

    def object_type(self):
        return self.property.type

    def item_title(self):
        return self.property.name

    def delete(self):
        self.context.del_property(self.property.id)


class PropertiedFormMixin(object):
    """ Mix in class for forms of :py:class:`horae.properties.interfaces.IPropertied`
    """
    # To be provided by subclass
    interfaces = (interfaces.IGlobalPropertiesHolder, interfaces.IGlobalProperties)
    initial = None
    editable = None
    display = True

    def completed(self, context):
        return interfaces.IComplete(context, lambda: False)()

    def offer(self, context):
        return interfaces.IOffer(context, lambda: False)()

    def property_fields(self, context=None):
        if context is None:
            context = self.context
        completed = self.completed(context)
        offer = self.offer(context)
        ifaces = [self.interfaces, PropertiedFormMixin.interfaces]
        fields = {}
        seen = []
        self.properties = []
        self.field_map = {}
        for iface_holder, iface in ifaces:
            holder = utils.findParentByInterface(context, iface_holder)
            if holder is None:
                continue
            container = iface(holder)
            properties = container.properties()
            for property in properties:
                if property.id in seen:
                    continue
                seen.append(property.id)
                if not self.display and ((not property.complete and completed) or (not property.offer and offer)):
                    continue
                self.properties.append(property)
                available = container.available(property, self.initial, self.editable, self.display)
                if available:
                    for field in property.fields(context):
                        if available:
                            fields[field.__name__] = field
                        self.field_map[field.__name__] = field
        return fields

    def setUpWidgets(self, ignore_request=False):
        super(PropertiedFormMixin, self).setUpWidgets(ignore_request)
        for property in self.properties:
            if property.remember:
                continue
            for field in property.field_names():
                if self.widgets.get(field) is None:
                    continue
                if not self.widgets[field].hasInput():
                    self.widgets[field].setRenderedValue(None)

    def validate(self, action, data):
        errors = super(PropertiedFormMixin, self).validate(action, data)
        for property in self.properties:
            errors += property.validate(data, self.context)
        return errors

    def update(self):
        fields = self.property_fields()
        super(PropertiedFormMixin, self).update()
        for form_field in self.form_fields:
            field = form_field.field
            fields[field.__name__] = field
        self.form_fields = grok.Fields(**fields)

    def new_change(self, obj):
        obj.new_change()

    def process_properties(self, **data):
        for property in self.properties:
            data = property.process(**data)
        return data

    def apply_properties(self, obj, **data):
        for property in self.properties:
            property.apply(obj, **data)


class PropertiedAddForm(PropertiedFormMixin, layout.AddForm):
    """ Add form for :py:class:`horae.properties.interfaces.IPropertied`
    """
    grok.baseclass()
    # To be provided by subclass
    factory = None
    container_interface = None
    initial = True
    editable = None
    display = None

    def completed(self, context):
        return False

    def apply(self, obj, **data):
        interfaces.IPropertied(obj).new_change()
        data = self.process_properties(**data)
        self.apply_properties(obj, **data)
        super(PropertiedAddForm, self).apply(obj, **data)

    def create(self, **data):
        obj = self.factory()
        return obj

    def add(self, obj):
        container = self.context
        while self.container_interface(container, None) is None:
            container = container.__parent__
        self.container_interface(container).add_object(obj)


class PropertiedEditForm(PropertiedFormMixin, layout.EditForm):
    """ Edit form for :py:class:`horae.properties.interfaces.IPropertied`
    """
    grok.baseclass()
    initial = None
    editable = True
    display = None

    def apply(self, **data):
        interfaces.IPropertied(self.context).new_change()
        data = self.process_properties(**data)
        self.apply_properties(self.context, **data)
        super(PropertiedEditForm, self).apply(**data)
        self.call_extenders('apply', obj=self.context, data=data)


class PropertyDisplayWidget(object):
    """ Display widget of a :py:class:`horae.properties.interfaces.IProperty`
    """

    def __init__(self, property, value, context, request, widget=None, cssClass=None):
        self._property = property
        self._value = value
        self._request = request
        self._widget = widget
        self._context = context
        self.prefix = widget is not None and widget._prefix or ''
        self.name = self.prefix + property.id
        self.label = property.name
        self.hint = property.description
        self.visible = True
        self.order = property.order
        self.cssClass = cssClass

    def __call__(self):
        if self._property.empty(self._value):
            return u''
        rendered = self._property.render(self._value, self._context, self._request, self._widget)
        if self.cssClass is not None:
            rendered = '<span class="%s">%s</span>' % (self.cssClass, rendered)
        return rendered


class PropertiedDisplayForm(PropertiedFormMixin, layout.DisplayForm):
    """ Display form for :py:class:`horae.properties.interfaces.IPropertied`
    """
    grok.baseclass()

    def setUpWidgets(self, ignore_request=False):
        super(PropertiedDisplayForm, self).setUpWidgets(ignore_request)
        property_map = dict([(property.id, property) for property in self.properties])
        new = []
        for widget in self.widgets:
            id = widget.name[len(widget._prefix):]
            if id in property_map:
                property = property_map[id]
                widget = PropertyDisplayWidget(property, interfaces.IPropertied(self.context).get_property(id), interfaces.IPropertied(self.context), self.request, widget)
            if not widget():
                continue
            new.append(widget)
        self.widgets = new
        providers = component.getAdapters((self.context,), interfaces.IPropertiedDisplayWidgetsProvider)
        for name, provider in providers:
            self.widgets = provider.widgets(self.widgets, self.request)
        self.widgets.sort(key=lambda x: getattr(x, 'order', 1))


class History(PropertiedDisplayForm):
    """ History of a :py:class:`horae.properties.interfaces.IPropertied`
    """
    grok.baseclass()
    grok.require('horae.ViewHistory')
    grok.template('history')
    template = ChameleonPageTemplate(os.path.join('templates', 'history.cpt'))
    grok.name('history')
    navigation.menuitem(IViewsMenu, _(u'History'))

    def __call__(self, change=None, previous=None, **kw):
        self.change = change
        self.previous = previous
        return super(History, self).__call__(**kw)

    def update(self):
        super(History, self).update()
        self.changes = []
        previous = None
        if self.change is None:
            previous = {}
            propertied = interfaces.IPropertied(self.context)
            changelog = propertied.changelog()
        else:
            changelog = [self.change, ]
        if self.previous is not None:
            previous = self.previous
        providers = [provider for name, provider in component.getAdapters((self.context,), interfaces.IHistoryPropertiesProvider)]
        for change in changelog:
            if previous is None:
                previous = dict(change.properties())
                continue
            creator = getUser(change.creator)
            item = {'id': change.id,
                    'user': creator is not None and creator.name or change.creator,
                    'user_url': creator is not None and IUserURL(creator, lambda: None)() or None,
                    'date': utils.formatDateTime(change.creation_date, self.request, ('dateTime', 'short')),
                    'name': change.get_property('name'),
                    'description': change.get_property('description'),
                    'title': change.get_property('title'),
                    'comment': change.get_property('comment'),
                    'workexpense': None,
                    'properties': []
                    }
            diff = {}
            for id, value in change.properties():
                field = id in self.field_map and self.field_map[id] or None
                if not field or not field.property.display:
                    continue
                if field and (not field.property.empty(value) or not field.property.empty(previous.get(id, None))):
                    if not field.property.id in diff:
                        diff[field.property.id] = 0.0
                    item['properties'].append({'name': field.title,
                                               'order': field.property.order + diff[field.property.id],
                                               'value': field.property.render(value, self.context, self.request),
                                               'previous': not field.property.empty(previous.get(id, None)) and field.property.render(previous.get(id, None), self.context, self.request) or None})
                    diff[field.property.id] += 0.1
            item['properties'].sort(key=lambda x: x['order'])
            for provider in providers:
                item['properties'] = provider.properties(change, item['properties'], self.request)
            previous.update(dict(change.properties()))
            if item['title'] or item['comment'] or item['name'] or item['description'] or item['properties']:
                self.changes.append(item)
        self.call_extenders('post_update')



from zope.formlib.itemswidgets import ItemDisplayWidget as BaseItemDisplayWidget

from horae.properties import _


class ItemDisplayWidget(BaseItemDisplayWidget):

    def __call__(self):
        """See IBrowserWidget."""
        value = self._getFormValue()
        if value is None or value == u'':
            return self.translate(self._messageNoValue)
        else:
            if not value in self.vocabulary:
                return self.translate(_(u'${value} (value no longer in vocabulary)', mapping=dict(value=value)))
            term = self.vocabulary.getTerm(value)
            return self.textForValue(term)



import grok
import re
from datetime import datetime, timedelta

from persistent import Persistent
from zope import schema
from zope import component
from zope import interface
from zope.i18n import translate
from zope.site.hooks import getSite
from zope.formlib.widget import renderElement, CustomWidgetFactory
from zope.schema.interfaces import ITitledTokenizedTerm
from zope.app.intid.interfaces import IIntIds
from zope.securitypolicy.interfaces import IPrincipalRoleManager

from z3c.relationfield.interfaces import IHasRelations
from z3c.relationfield import RelationValue
from megrok.form import fields

from horae.core import utils
from horae.core.interfaces import ICurrencyFormatter
from horae.auth.utils import getUser, getGroup
from horae.auth.interfaces import IUserURL
from horae.layout.widgets import ObjectWidget
from horae.layout.widgets import ListSequenceWidget
from horae.timeaware.timeaware import DatetimeFieldProperty
from horae.autocomplete.fields import AutocompleteChoice

from horae.properties import _
from horae.properties import interfaces


class PropertyTypes(object):
    """ Provider for available properties
    """
    grok.implements(interfaces.IPropertyTypes)

    def __init__(self):
        self._types = []

    def add(self, type):
        if not interfaces.IProperty:
            raise ValueError(type)
        self._types.append(type)

    def types(self):
        """ Returns a list of available property types
        """
        return self._types

property_types = PropertyTypes()


def property_types_singleton():
    """ Returns the :py:class:`PropertyTypes` singleton
    """
    return property_types
grok.global_utility(property_types_singleton, interfaces.IPropertyTypes)


class Property(grok.Model):
    """ A definition of a property
    """
    grok.baseclass()

    id = schema.fieldproperty.FieldProperty(interfaces.IProperty['id'])
    name = schema.fieldproperty.FieldProperty(interfaces.IProperty['name'])
    description = schema.fieldproperty.FieldProperty(interfaces.IProperty['description'])
    required = schema.fieldproperty.FieldProperty(interfaces.IProperty['required'])
    initial = schema.fieldproperty.FieldProperty(interfaces.IProperty['initial'])
    editable = schema.fieldproperty.FieldProperty(interfaces.IProperty['editable'])
    order = schema.fieldproperty.FieldProperty(interfaces.IProperty['order'])
    display = schema.fieldproperty.FieldProperty(interfaces.IProperty['display'])
    remember = schema.fieldproperty.FieldProperty(interfaces.IProperty['remember'])
    searchable = schema.fieldproperty.FieldProperty(interfaces.IProperty['searchable'])
    complete = schema.fieldproperty.FieldProperty(interfaces.IProperty['complete'])
    offer = schema.fieldproperty.FieldProperty(interfaces.IProperty['offer'])

    customizable = True
    permission = None

    def _prepare_fields(self, fields):
        i = 0.0
        for field in fields:
            field.interface = interfaces.IPropertiesProxy
            field.property = self
            field.order = self.order + i
            interface.alsoProvides(field, interfaces.IPropertyField)
            i += 0.0001
        return fields

    def custom_widget(self, form_fields):
        """ Called after the form fields for the add or edit form of a property have been created
            may be used to add custom widgets to specific fields
        """
        pass

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        raise NotImplementedError(u'concrete classes must implement fields()')

    def field_names(self):
        """ Returns a list of field names provided by this property
        """
        return [self.id, ]

    def process(self, **data):
        """ Processes the input coming from the form and returns a new data dictionary
        """
        return data

    def apply(self, obj, **data):
        """ Applies the data to the obj
        """
        pass

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        return widget is not None and widget() or value

    def clone(self):
        """ Returns a copy of the property
        """
        clone = self.__class__.__new__(self.__class__)
        clone.__dict__.update(self.__dict__)
        clone.id = self.id
        clone.__name__ = clone.__parent__ = None
        return clone

    def validate(self, data, context):
        """ Validates the form data

            Returns a list of exceptions or an empty one if the data is valid
        """
        return []

    def empty(self, value):
        """ Decides whether the value provided is an empty value for this property type
        """
        return value is None


class BoolProperty(Property):
    """ A boolean property
    """
    grok.implements(interfaces.IBoolProperty)
    type = _(u'Boolean')
    default = schema.fieldproperty.FieldProperty(interfaces.IBoolProperty['default'])

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        return value and _(u'Yes') or _(u'No')

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            schema.Bool(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default
            ),
        ])
property_types.add(BoolProperty)


class TextLineProperty(Property):
    """ A textline property
    """
    grok.implements(interfaces.ITextLineProperty)
    type = _(u'Text line')
    default = schema.fieldproperty.FieldProperty(interfaces.ITextLineProperty['default'])

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            schema.TextLine(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default
            ),
        ])
property_types.add(TextLineProperty)


class TextProperty(Property):
    """ A text property
    """
    grok.implements(interfaces.ITextProperty)
    type = _(u'Text')
    default = schema.fieldproperty.FieldProperty(interfaces.ITextProperty['default'])

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            schema.Text(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default
            ),
        ])
property_types.add(TextProperty)


class RichTextProperty(Property):
    """ A rich text property
    """
    grok.implements(interfaces.IRichTextProperty)
    type = _(u'Rich text')
    default = u''

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            fields.HTML(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default
            ),
        ])
property_types.add(RichTextProperty)


class ChoiceProperty(Property):
    """ A choice property
    """
    grok.implements(interfaces.IChoiceProperty)
    type = _(u'Choice')
    vocabulary = schema.fieldproperty.FieldProperty(interfaces.IChoiceProperty['vocabulary'])
    default = schema.fieldproperty.FieldProperty(interfaces.IChoiceProperty['default'])
    numbered = re.compile(r'^[0-9]+')

    def convertValue(self, value):
        numbered = self.numbered.match(value)
        if numbered is None:
            return value
        return int(numbered.group(0))

    def getVocabulary(self, context):
        """ Returns the vocabulary used by the field
        """
        terms = []
        for value in self.vocabulary:
            converted = self.convertValue(value)
            terms.append(schema.vocabulary.SimpleTerm(converted, converted, value))
        return schema.vocabulary.SimpleVocabulary(terms)

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if widget is not None:
            return widget()
        if not value:
            return value
        vocabulary = self.getVocabulary(context)
        if not value in vocabulary:
            return translate(_(u'${value} (value no longer in vocabulary)', mapping=dict(value=value)), context=request)
        term = vocabulary.getTerm(value)
        if ITitledTokenizedTerm.providedBy(term):
            return term.title
        return term.token

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            schema.Choice(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.convertValue(self.default),
                vocabulary=self.getVocabulary(context)
            ),
        ])
property_types.add(ChoiceProperty)


class MultipleChoiceProperty(ChoiceProperty):
    """ A multiple choice property
    """

    def convertValue(self, value):
        converted = set([])
        if not value:
            return converted
        for val in value:
            converted.add(super(MultipleChoiceProperty, self).convertValue(val))
        return converted

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if widget is not None:
            return widget()
        if not value:
            return value
        return ', '.join([super(MultipleChoiceProperty, self).render(val, context, request, widget) for val in value])

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        vocab = self.getVocabulary(context)
        if not len(vocab):
            return []
        return self._prepare_fields([
            schema.Set(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.convertValue(self.default),
                value_type=schema.Choice(
                    vocabulary=vocab
                )
            ),
        ])
property_types.add(MultipleChoiceProperty)


class WeightedOption(Persistent):
    """ A weighted option
    """
    grok.implements(interfaces.IWeightedOption)
    name = schema.fieldproperty.FieldProperty(interfaces.IWeightedOption['name'])
    affects_sorting = schema.fieldproperty.FieldProperty(interfaces.IWeightedOption['weight'])

    def __init__(self, name=u'', weight=1):
        super(WeightedOption, self).__init__()
        self.name = name
        self.weight = weight


class WeightedChoiceProperty(ChoiceProperty):
    """ A choice property having weighted options
    """
    grok.implements(interfaces.IWeightedChoiceProperty)
    type = _(u'Weighted choice')
    vocabulary = schema.fieldproperty.FieldProperty(interfaces.IWeightedChoiceProperty['vocabulary'])
    affects_sorting = schema.fieldproperty.FieldProperty(interfaces.IWeightedChoiceProperty['affects_sorting'])

    def custom_widget(self, form_fields):
        """ Called after the form fields for the add or edit form of a property have been created
            may be used to add custom widgets to specific fields
        """
        form_fields['vocabulary'].custom_widget = CustomWidgetFactory(ListSequenceWidget, subwidget=CustomWidgetFactory(ObjectWidget, WeightedOption))

    def getVocabulary(self, context):
        """ Returns the vocabulary used by the field
        """
        terms = []
        for value in self.vocabulary:
            terms.append(schema.vocabulary.SimpleTerm(value.name, value.name, value.name))
        return schema.vocabulary.SimpleVocabulary(terms)
property_types.add(WeightedChoiceProperty)


class FloatProperty(Property):
    """ A float property
    """
    grok.implements(interfaces.IFloatProperty)
    type = _(u'Float')
    default = schema.fieldproperty.FieldProperty(interfaces.IFloatProperty['default'])
    min = schema.fieldproperty.FieldProperty(interfaces.IFloatProperty['min'])
    max = schema.fieldproperty.FieldProperty(interfaces.IFloatProperty['max'])

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if value is None:
            return value
        return utils.formatHours(value, request)

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            schema.Float(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default,
                min=self.min,
                max=self.max
            ),
        ])
property_types.add(FloatProperty)


class PriceProperty(FloatProperty):
    """ A price property
    """
    grok.implements(interfaces.IPriceProperty)
    type = _(u'Price')

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if value is None:
            return value
        return component.getMultiAdapter((getSite(), request), interface=ICurrencyFormatter).format(value)
property_types.add(PriceProperty)


class UserProperty(Property):
    """ A user property
    """
    grok.implements(interfaces.IUserProperty)
    type = _(u'User')
    default = None

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if value is None:
            return value
        user = getUser(value)
        url = user and component.queryAdapter(user, interface=IUserURL) or None
        return user and (url and '<a href="%s">%s</a>' % (url(), user.name) or user.name) or value

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            AutocompleteChoice(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default,
                vocabulary='horae.auth.vocabulary.usernames'
            ),
        ])
property_types.add(UserProperty)


class UserRoleProperty(UserProperty):
    """ A user role property
    """
    grok.implements(interfaces.IUserRoleProperty)
    type = _(u'User role')
    role = None

    def apply(self, obj, **data):
        """ Applies the data to the obj
        """
        manager = IPrincipalRoleManager(obj)
        principals = manager.getPrincipalsForRole(self.role)
        for principal, setting in principals:
            manager.unsetRoleForPrincipal(self.role, principal)
        if self.id in data and data[self.id]:
            manager.assignRoleToPrincipal(self.role, data[self.id])
property_types.add(UserRoleProperty)


class GroupProperty(Property):
    """ A group property
    """
    grok.implements(interfaces.IGroupProperty)
    type = _(u'Group')
    default = None

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if value is None:
            return value
        user = getGroup(value)
        return user and user.name or value

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            AutocompleteChoice(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default,
                vocabulary='horae.auth.vocabulary.groupids'
            ),
        ])
property_types.add(GroupProperty)


class GroupRoleProperty(GroupProperty):
    """ A group role property
    """
    grok.implements(interfaces.IGroupRoleProperty)
    type = _(u'Group role')
    role = None

    def apply(self, obj, **data):
        """ Applies the data to the obj
        """
        manager = IPrincipalRoleManager(obj)
        principals = manager.getPrincipalsForRole(self.role)
        for principal, setting in principals:
            manager.unsetRoleForPrincipal(self.role, principal)
        if self.id in data and data[self.id]:
            manager.assignRoleToPrincipal(self.role, data[self.id])
property_types.add(GroupRoleProperty)


class DateTimeProperty(Property):
    """ A date time property
    """
    grok.implements(interfaces.IDateTimeProperty)
    type = _(u'Datetime')
    default = DatetimeFieldProperty(interfaces.IDateTimeProperty['default'])
    default_now = schema.fieldproperty.FieldProperty(interfaces.IDateTimeProperty['default_now'])

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            schema.Datetime(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default_now and datetime.now() or self.default
            ),
        ])

    def process(self, **data):
        """ Processes the input coming from the form and returns a new data dictionary
        """
        if self.id in data and isinstance(data[self.id], datetime):
            data[self.id] = datetime(data[self.id].year, data[self.id].month, data[self.id].day, data[self.id].hour, data[self.id].minute)
        return data

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if not isinstance(value, datetime):
            return value
        if value:
            value = utils.formatDateTime(value, request)
        return value
property_types.add(DateTimeProperty)


class DateTimeRangeProperty(DateTimeProperty):
    """ A date time range property
    """
    grok.implements(interfaces.IDateTimeRangeProperty)
    type = _(u'Datetime range')
    name_end = schema.fieldproperty.FieldProperty(interfaces.IDateTimeRangeProperty['name_end'])
    description_end = schema.fieldproperty.FieldProperty(interfaces.IDateTimeRangeProperty['description_end'])
    required_end = schema.fieldproperty.FieldProperty(interfaces.IDateTimeRangeProperty['required_end'])
    default_diff = schema.fieldproperty.FieldProperty(interfaces.IDateTimeRangeProperty['default_diff'])
    hours = schema.fieldproperty.FieldProperty(interfaces.IDateTimeRangeProperty['hours'])

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        default_end = self.default_now and datetime.now() or self.default
        default_start = default_end is not None and default_end - timedelta(hours=float(self.default_diff)) or None
        fields = []
        if self.hours:
            fields.append(
                schema.Float(
                    __name__=self.id + '_hours',
                    title=_(u'Hours'),
                    required=False,
                    default=float(self.default_diff)
                )
            )
        fields.extend([
            schema.Datetime(
                __name__=self.id + '_start',
                title=self.name,
                description=self.description,
                required=self.required,
                default=default_start
            ),
            schema.Datetime(
                __name__=self.id + '_end',
                title=self.name_end,
                description=self.description_end,
                required=self.required_end,
                default=default_end
            ),
        ])
        return self._prepare_fields(fields)

    def field_names(self):
        """ Returns a list of field names provided by this property
        """
        return [self.id + '_start', self.id + '_end', ]

    def process(self, **data):
        """ Processes the input coming from the form and returns a new data dictionary
        """
        if self.id + '_hours' in data:
            del data[self.id + '_hours']
        if self.id + '_start' in data and isinstance(data[self.id + '_start'], datetime):
            data[self.id + '_start'] = datetime(data[self.id + '_start'].year, data[self.id + '_start'].month, data[self.id + '_start'].day, data[self.id + '_start'].hour, data[self.id + '_start'].minute)
        if self.id + '_end' in data and isinstance(data[self.id + '_end'], datetime):
            data[self.id + '_end'] = datetime(data[self.id + '_end'].year, data[self.id + '_end'].month, data[self.id + '_end'].day, data[self.id + '_end'].hour, data[self.id + '_end'].minute)
        return data

    def validate(self, data, context):
        """ Validates the form data

            Returns a list of exceptions or an empty one if the data is valid
        """
        if self.id + '_start' in data and \
           self.id + '_end' in data and \
           data[self.id + '_start'] is not None and \
           data[self.id + '_end'] is not None and \
           data[self.id + '_start'] > data[self.id + '_end']:
            return [interface.Invalid(_(u'${end_date} before ${start_date}', mapping={'end_date': self.name_end, 'start_date': self.name})), ]
        return []
property_types.add(DateTimeRangeProperty)


class MilestoneProperty(Property):
    """ A milestone property
    """
    grok.implements(interfaces.IMilestoneProperty)
    type = _(u'Milestone')
    default = schema.fieldproperty.FieldProperty(interfaces.IMilestoneProperty['default'])

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            schema.Choice(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=self.default,
                vocabulary='horae.properties.vocabulary.milestones'
            ),
        ])

    def apply(self, obj, **data):
        """ Applies the data to the obj
        """
        if not self.id in data or \
           not IHasRelations.providedBy(obj) or \
           not hasattr(obj, self.id + '_rel'):
            return
        milestone = data[self.id]
        if milestone is None:
            setattr(obj, self.id + '_rel', None)
            return
        intids = component.getUtility(IIntIds)
        setattr(obj, self.id + '_rel', RelationValue(intids.queryId(milestone)))

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if value is None and widget:
            if widget._renderedValueSet():
                value = widget._data
        if value is None:
            return super(MilestoneProperty, self).render(value, request, widget)
        return renderElement('a',
                             href=component.getMultiAdapter((value, request), name='absolute_url'),
                             title=value.description,
                             contents=widget is not None and widget() or value.name)


class FieldsProperty(Property):
    """ A fields property
    """
    grok.implements(interfaces.IFieldsProperty)
    type = _(u'Fields')

    def fields(self, context):
        """ Returns a list of fields (zope.formlib.interfaces.IField)
        """
        return self._prepare_fields([
            schema.Set(
                __name__=self.id,
                title=self.name,
                description=self.description,
                required=self.required,
                default=set(),
                value_type=schema.Choice(
                    vocabulary='horae.properties.vocabulary.fields'
                )
            ),
        ])

    def render(self, value, context, request, widget=None):
        """ Returns the rendered value
        """
        if widget is not None:
            return widget()
        if value is None or not len(value):
            return None
        return ', '.join([v for v in iter(value)])



import grok

from zope import component, schema
from zope.security import checkPermission

from BTrees.OOBTree import OOBTree

from horae.core.interfaces import IIntIdManager
from horae.lifecycle import lifecycle
from horae.cache import request

from horae.properties import interfaces


class PropertyChange(grok.Model, lifecycle.LifecycleAwareMixin):
    """ A change of multiple properties
    """
    grok.implements(interfaces.IPropertyChange)
    id = schema.fieldproperty.FieldProperty(interfaces.IPropertyChange['id'])

    def __init__(self):
        super(PropertyChange, self).__init__()
        self._properties = OOBTree()

    def properties(self):
        """ Iterator over the changed properties (name, value)
        """
        properties = self._properties.items()
        hidden = []
        if not checkPermission('horae.ViewHiddenProperties', self):
            hidden = self._properties.get('hidden', [])
        if not 'ALL' in hidden:
            for id, value in properties:
                if not id in hidden:
                    yield id, value

    def get_property(self, id, default=None):
        """ Return the specific property value
        """
        hidden = self._properties.get('hidden', [])
        if ((id in hidden or 'ALL' in hidden) and
            not checkPermission('horae.ViewHiddenProperties', self)):
            return default
        return self._properties.get(id, default)

    def set_property(self, id, value):
        """ Sets a property
        """
        self._properties[id] = value


class PropertiedMixin(object):
    """ An object storing dynamic properties and providing a history over them
    """
    grok.implements(interfaces.IPropertied)

    def __init__(self):
        super(PropertiedMixin, self).__init__()
        self._current = None

    @property
    def _properties(self):
        if not hasattr(self, '_props'):
            self._props = OOBTree()
        return self._props

    def _changes(self):
        if not 'changes' in self:
            self['changes'] = grok.OrderedContainer()
        return self['changes']

    def properties(self):
        """ Iterator over the available properties (IProperty)
        """
        providers = reversed([provider for name, provider in component.getAdapters((self,), interfaces.IProperties) if not name == ''])
        seen = []
        for provider in providers:
            for property in provider.properties():
                if property.id in seen:
                    continue
                seen.append(property.id)
                yield property

    @request.cache()
    def find_property(self, id):
        """ Trys to find and return the property having the specified id
        """
        providers = reversed([provider for name, provider in component.getAdapters((self,), interfaces.IProperties) if not name == ''])
        for provider in providers:
            property = provider.get_property(id)
            if property is not None:
                return property
        return None

    def new_change(self):
        """ Starts a new property change and returns it
        """
        self._current = PropertyChange()
        self._current.id = component.getUtility(IIntIdManager).nextId('property_change')
        order = [str(self._current.id), ] + list(self._changes().keys())
        self._changes()[str(self._current.id)] = self._current
        self._changes().updateOrder(order)
        return self._current

    def get_property(self, id, default=None):
        """ Return the specific property value
        """
        value = None
        if id in self._properties:
            value = self._properties[id]
        else:
            for change in self._changes().values():
                value = change.get_property(id)
                if value is not None:
                    self._properties[id] = value
                    break
        if value is None:
            return default
        return value

    def set_property(self, id, value):
        """ Sets a property on the currently active property change
        """
        self._current.set_property(id, value)
        self._properties[id] = value

    def changelog(self):
        """ Iterator over the history of property changes
        """
        hidden = checkPermission('horae.ViewHiddenProperties', self)
        for change in self._changes().values():
            if not hidden and 'ALL' in change.get_property('hidden', []):
                continue
            yield change

    def current(self):
        """ Returns the current property change
        """
        return self._current

    def __getattr__(self, name):
        if name.startswith('_'):
            return object.__getattribute__(self, name)
        if name in self._properties:
            return self._properties[name]
        for property in self.properties():
            if property.remember and name in property.field_names():
                return self.get_property(name)
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return None


class TicketPropertiedMixin(PropertiedMixin):
    """ Mix in class for ticket propertied
    """

    def properties(self):
        """ Iterator over the available properties (IProperty)
        """
        providers = reversed([provider for name, provider in component.getAdapters((self,), interfaces.IProperties) if not name == ''])
        _marker = object()
        milestone = self.get_property('milestone', _marker)
        if milestone is _marker:
            for provider in providers:
                property = provider.get_property('milestone')
                if property:
                    milestone = property.default
                    break
        if milestone is not None:
            providers = [interfaces.ITicketProperties(milestone), ] + [provider for provider in providers]
        seen = []
        for provider in providers:
            for property in provider.properties():
                if property.id in seen:
                    continue
                seen.append(property.id)
                yield property


class PropertiesProxy(grok.Adapter):
    """ Proxy around IPropertied
    """
    grok.context(interfaces.IPropertied)
    grok.implements(interfaces.IPropertiesProxy)

    def __init__(self, propertied):
        self._propertied = propertied

    def __getattribute__(self, name):
        """ Returns the current property value
        """
        return getattr(object.__getattribute__(self, '_propertied'), name)

    def __setattr__(self, name, value):
        """ Sets the value to the last property change
        """
        if name == '_propertied':
            object.__setattr__(self, name, value)
        else:
            object.__getattribute__(self, '_propertied').set_property(name, value)


@grok.adapter(PropertiesProxy)
@grok.implementer(interfaces.IPropertied)
def propertied_of_properties_proxy(proxy):
    """ Returns the propertied of a :py:class:`PropertiesProxy`
    """
    return object.__getattribute__(proxy, '_propertied')



from zope.schema import vocabulary

from horae.properties import _
from horae.properties import interfaces


def fields_vocabulary_factory(context):
    """ A vocabulary of all available fields registered as
        **horae.properties.vocabulary.fields**
    """
    terms = [vocabulary.SimpleTerm('ALL', 'ALL', _(u'All fields'))]
    for property in interfaces.IPropertied(context).properties():
        for field in property.fields(context):
            if not interfaces.IFieldsProperty.providedBy(field.property):
                terms.append(vocabulary.SimpleTerm(field.__name__, field.__name__, field.title))
    return vocabulary.SimpleVocabulary(terms)
vocabulary.getVocabularyRegistry().register('horae.properties.vocabulary.fields', fields_vocabulary_factory)
